# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import setpath
setpath.set_path()
import ikarus as iks
import ikarus.finite_elements
import ikarus.utils
import ikarus.assembler
import ikarus.dirichletValues
import numpy as np
import scipy as sp
from scipy.optimize import minimize

import dune.grid
import dune.functions
from dune.vtk import vtkUnstructuredGridWriter, vtkWriter, RangeTypes, FieldInfo

if __name__ == "__main__":

    # help(iks)
    lowerLeft = []
    upperRight = []
    elements = []
    for i in range(2):
        lowerLeft.append(-1)
        upperRight.append(1)
        elements.append(3)

    req= ikarus.FErequirements()
    req.addAffordance(iks.ScalarAffordances.mechanicalPotentialEnergy)

    grid = dune.grid.structuredGrid(lowerLeft,upperRight,elements)
    grid. hierarchicalGrid.globalRefine(0)
    basisLagrange1 = ikarus.basis(grid, dune.functions.Power(dune.functions.Lagrange(order=1),2))
    flatBasis = basisLagrange1.flat()


    forces = np.zeros(8)
    stiffness = np.zeros((8,8))

    def volumeLoad(x,lambdaVal) :
        return np.array([lambdaVal*x[0]*2, 2*lambdaVal*x[1]*0])

    def neumannLoad(x,lambdaVal) :
        return np.array([lambdaVal*0, lambdaVal])


    neumannVertices = np.zeros(grid.size(2)*2, dtype=bool)
    assert len(neumannVertices)== len(flatBasis)

    flatBasis.interpolate(neumannVertices, lambda x :  True  if x[1]>0.9 else False)

    boundaryPatch = iks.utils.boundaryPatch(grid,neumannVertices)

    svk = iks.StVenantKirchhoff(emodul=1000,nu=0.3)

    psNH = svk.asPlaneStress()
    fes = []
    for e in grid.elements:
        fes.append(iks.finite_elements.nonLinearElasticElement(basisLagrange1,e,psNH,volumeLoad,boundaryPatch,neumannLoad))

    dirichletValues = iks.dirichletValues(flatBasis)

    def fixFirstIndex(vec,globalIndex):
        vec[0]= True

    def fixAnotherVertex(vec,localIndex,localView):
        localView.index(localIndex)
        vec[1]= True

    def fixLeftHandEdge(vec,localIndex,localView,intersection):
        if (intersection.geometry.center[1]<-0.9):
            vec[localView.index(localIndex)]= True

    dirichletValues.fixBoundaryDOFs(fixFirstIndex)
    dirichletValues.fixBoundaryDOFsUsingLocalView(fixAnotherVertex)
    dirichletValues.fixBoundaryDOFsUsingLocalViewAndIntersection(fixLeftHandEdge)

    assembler = iks.assembler.sparseFlatAssembler(fes,dirichletValues)

    dRed = np.zeros(assembler.reducedSize())

    lambdaLoad = iks.ValueWrapper(3.0)
    def energy(dRedInput) :
        global assembler

        reqL= ikarus.FErequirements()
        reqL.addAffordance(iks.ScalarAffordances.mechanicalPotentialEnergy)
        reqL.insertParameter(iks.FEParameter.loadfactor,lambdaLoad)

        dBig = assembler.createFullVector(dRedInput)
        reqL.insertGlobalSolution(iks.FESolutions.displacement,dBig)
        return assembler.getScalar(reqL)
    def gradient(dRedInput) :
        global assembler

        reqL= ikarus.FErequirements()
        reqL.addAffordance(iks.VectorAffordances.forces)
        reqL.insertParameter(iks.FEParameter.loadfactor,lambdaLoad)

        dBig = assembler.createFullVector(dRedInput)
        reqL.insertGlobalSolution(iks.FESolutions.displacement,dBig)
        return assembler.getReducedVector(reqL)
    def hess(dRedInput) :
        global assembler

        reqL= ikarus.FErequirements()
        reqL.addAffordance(iks.MatrixAffordances.stiffness)
        reqL.insertParameter(iks.FEParameter.loadfactor,lambdaLoad)

        dBig = assembler.createFullVector(dRedInput)
        reqL.insertGlobalSolution(iks.FESolutions.displacement,dBig)
        return assembler.getReducedMatrix(reqL).todense()

    print("energy(dRed):",energy(dRed))
    print("energyafer")
    resultd = minimize(energy,x0=dRed,options={"disp": True},tol=1e-14)
    resultd2 = minimize(energy,x0=dRed,jac=gradient,options={"disp": True},tol=1e-14)
    resultd3 = minimize(energy,method="trust-constr",x0=dRed,jac=gradient,hess=hess,options={ 'disp': True})
    resultd4 = sp.optimize.root(gradient,jac=hess,x0=dRed,options={ 'disp': True},tol=1e-10)
    # print(assembler.createFullVector(resultd.g))
    np.set_printoptions(precision=3)

    assert(np.allclose(resultd.x,resultd2.x,atol=1e-6))
    assert(np.allclose(resultd3.x,resultd4.x))
    assert(np.all(abs(resultd3.grad)<1e-8))
    assert(np.all(abs(resultd4.fun)<1e-8))



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.generator.generator import SimpleGenerator
from dune.common.hashit import hashIt
import dune

def decoratePre(pre):
    def wrappedPre(*args, **kwargs):
        preamble = pre(*args, **kwargs)
        newPreamble = ""
        newPreamble += "#define DUNE_LOCALFEFUNCTIONS_USE_EIGEN 1\n"
        newPreamble += "#define EIGEN_DEFAULT_TO_ROW_MAJOR 1\n" #needed to have conforming Matrix storage between eigen and numpy otherwise references are not working
        newPreamble += preamble
        return newPreamble

    return wrappedPre


myAttributes = vars(SimpleGenerator).copy()
myAttributes["pre"] = decoratePre(myAttributes["pre"])
MySimpleGenerator = type("MySimpleGenerator", (object,), myAttributes)


def sparseFlatAssembler(fes, dirichletValues) :
    generator = MySimpleGenerator("SparseFlatAssembler", "Ikarus::Python")
    element_type = f"Ikarus::SparseFlatAssembler<std::vector<{fes[0].cppTypeName}>,{dirichletValues.cppTypeName}>"

    includes = []
    includes += ["ikarus/assembler/simpleAssemblers.hh"]
    includes += fes[0]._includes # include header of finite element
    includes += ["ikarus/python/assembler/flatAssembler.hh"]
    moduleName = "SparseFlatAssembler_" + hashIt(element_type)
    module = generator.load(
        includes=includes,
        typeName=element_type,
        moduleName=moduleName
    )
    return module.SparseFlatAssembler(fes,dirichletValues)


def denseFlatAssembler(fes, dirichletValues) :
    # dune.generator.addToFlags(pre="-DCMAKE_PREFIX_PATH=/dune/dune-fufem ")
    generator = MySimpleGenerator("DenseFlatAssembler", "Ikarus::Python")
    element_type = f"Ikarus::DenseFlatAssembler<std::vector<{fes[0].cppTypeName}>,{dirichletValues.cppTypeName}>"

    includes = []
    includes += ["ikarus/assembler/simpleAssemblers.hh"]
    includes += ["ikarus/finiteElements/mechanics/linearElastic.hh"]
    includes += ["ikarus/python/assembler/flatAssembler.hh"]
    moduleName = "SparseFlatAssembler_" + hashIt(element_type)
    module = generator.load(
        includes=includes,
        typeName=element_type,
        moduleName=moduleName
    )
    return module.DenseFlatAssembler(fes,dirichletValues)



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.generator.generator import SimpleGenerator
from dune.common.hashit import hashIt
import dune

def decoratePre(pre):
    def wrappedPre(*args, **kwargs):
        preamble = pre(*args, **kwargs)
        newPreamble = ""
        newPreamble += "#define DUNE_LOCALFEFUNCTIONS_USE_EIGEN 1\n"
        newPreamble += "#define EIGEN_DEFAULT_TO_ROW_MAJOR 1\n" #needed to have conforming Matrix storage between eigen and numpy otherwise references are not working
        newPreamble += preamble
        return newPreamble

    return wrappedPre


myAttributes = vars(SimpleGenerator).copy()
myAttributes["pre"] = decoratePre(myAttributes["pre"])
MySimpleGenerator = type("MySimpleGenerator", (object,), myAttributes)


def linearElasticElement(basis, element, youngsMod, nu,volumeLoad=None,bp=None,neumannBoundaryLoad=None) :
    if not ((bp is None and neumannBoundaryLoad is None) or (bp is not None and neumannBoundaryLoad is not None)):
        raise TypeError("If you provide a boundary patch you should also provide a boundary load!")

    generator = MySimpleGenerator("LinearElastic", "Ikarus::Python")
    element_type = f"Ikarus::LinearElastic<{basis.cppTypeName},Ikarus::FErequirements<Eigen::Ref<Eigen::VectorXd>>,true>"

    includes = []
    includes += ["ikarus/finiteElements/mechanics/linearElastic.hh"]
    includes += ["ikarus/python/finiteElements/linearElastic.hh"]
    moduleName = "linearElastic_" + hashIt(element_type)
    module = generator.load(
        includes=includes,
        typeName=element_type,
        moduleName=moduleName
    )
    # https://pybind11.readthedocs.io/en/stable/advanced/functions.html#allow-prohibiting-none-arguments
    if volumeLoad is None:
        return module.LinearElastic(basis, element, youngsMod, nu)
    elif (bp is None and neumannBoundaryLoad is None):
        return module.LinearElastic(basis, element, youngsMod, nu,volumeLoad)
    else :
        return module.LinearElastic(basis, element, youngsMod, nu,volumeLoad,bp,neumannBoundaryLoad)


def nonLinearElasticElement(basis, element, material,volumeLoad=None,bp=None,neumannBoundaryLoad=None) :
    if not ((bp is None and neumannBoundaryLoad is None) or (bp is not None and neumannBoundaryLoad is not None)):
        raise TypeError("If you provide a boundary patch you should also provide a boundary load!")

    generator = MySimpleGenerator("NonLinearElastic", "Ikarus::Python")
    element_type = f"Ikarus::NonLinearElastic<{basis.cppTypeName},  {material.cppTypeName} ,Ikarus::FErequirements<Eigen::Ref<Eigen::VectorXd>>,true>"

    includes = []
    includes += ["ikarus/finiteElements/mechanics/nonLinearElastic.hh"]
    includes += ["ikarus/python/finiteElements/nonLinearElastic.hh"]
    moduleName = "nonLinearElastic_" + hashIt(element_type)
    module = generator.load(
        includes=includes,
        typeName=element_type,
        moduleName=moduleName
    )
    # https://pybind11.readthedocs.io/en/stable/advanced/functions.html#allow-prohibiting-none-arguments
    if volumeLoad is None:
        return module.NonLinearElastic(basis, element, material)
    elif bp is None and neumannBoundaryLoad is None:
        return module.NonLinearElastic(basis, element, material,volumeLoad)
    else :
        return module.NonLinearElastic(basis, element, material,volumeLoad,bp,neumannBoundaryLoad)
